/***************************************************************************
 *   Copyright (C) 2010 by Veli Mkinen                                    *
 *                                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
 ***************************************************************************/
 
// Modified by Leena Salmela 2011

#include <sstream>
#include <iostream>
using std::cout;
using std::endl;
using std::cerr;
#include <string>
using std::string;
#include <ctime>
#include <cstring>
#include <getopt.h>
#include <math.h>
#include "InputReader.h"
#include "Pattern.h"

#ifdef PARALLEL_SUPPORT
#include <omp.h>
#endif


std::pair<unsigned,unsigned> overlapAlign(string A, Pattern P, unsigned maxindel, unsigned minaveragequality, unsigned minprefixoverlap, unsigned minsuffixoverlap, unsigned mismatchlimit)
{
   // returns <px,sx> where px (sx) is largest overlap of 
   // suffix (prefix) of A and prefix (suffix) of P with at most maxindel indels,
   // average quality scores of matches in the alignment is at least 
   // minaveragequality, px>=minprefixoverlap, and sx>=minsuffixoverlap

   // D[i,j,k]=<jj,s>="jj such that A[jj-1,j-1] overlaps with P[0,i-1] with k indels maximizing average score s"
   unsigned px;
   unsigned sx;
   // keeping only two columns of D
   std::pair<int,int> **Dprev = new std::pair<int,int>*[P.size()+1]; 
   std::pair<int,int> **D = new std::pair<int,int>*[P.size()+1];        
   std::pair<int,int> infinitypair = std::make_pair(A.size(), -255*A.size());
   // initializing first column, and first row of second column
   for (unsigned i=0; i<=P.size();i++) {
       Dprev[i] = new std::pair<int,int>[maxindel+1];
       D[i] = new std::pair<int,int>[maxindel+1];
       for (unsigned k=0; k<=maxindel;k++) {
           if (i==k)  
              Dprev[i][k] = std::make_pair(0,0);
           else
             Dprev[i][k] = std::make_pair(0,0);
           //              Dprev[i][k] = infinitypair;
           if (i==0)
              D[i][k] = std::make_pair(0,0);
       }
   }
    
   // filling "matrix" column-by-column
   for (unsigned j=1; j<=A.size(); j++) {
      for (unsigned i=1; i<=P.size(); i++)
         for (unsigned k=0; k<=maxindel; k++) {
            std::pair<int,int> left = infinitypair, diag = infinitypair, up = infinitypair;
            if (k>0) {
               left = Dprev[i][k-1];
               up = D[i-1][k-1];
            }
            // if P[i-1]==A[j-1] with probability (1 - error probability p) then matchscore = -10 log_10 p 
            // if P[i-1]!=A[j-1] then match probability is q=p/3 and matchscore = -10 log_10 (1-q)
#ifdef IGNORE_QUALITY
            int matchscore = 40;
#else
            int matchscore = int(P.getQuality()[i-1])-64.0; // from ascii to score          
#endif
            double p = pow(10.0, -matchscore/10.0);
            double q = p/3; 
            int mismatchscore = -10*log10(1-q);
            int diagonalscore = (P.getPattern()[i-1]==A[j-1]?matchscore:mismatchscore);

	    if (matchscore > (int)mismatchlimit) {
   	       if (P.getPattern()[i-1]==A[j-1]) 
                  diag = Dprev[i-1][k];   
	       else if (k>0)
                  diag = Dprev[i-1][k-1];   	       
	    }
	    else 
  	       diag = Dprev[i-1][k];   	    
            // take max of three scores
            int jj = left.first;
            int s = left.second;
            if (s<up.second) {
               jj = up.first;
               s = up.second;
            }
            if (s<diag.second+diagonalscore) {
               jj = diag.first;
               s = diag.second+diagonalscore;
            } 
            D[i][k] = std::make_pair(jj,s); 
         }         
      // copy D to Dprev
      for (unsigned i=1; i<=P.size(); i++)
         for (unsigned k=0; k<=maxindel; k++) 
            Dprev[i][k]=D[i][k];
      Dprev[0][0]=std::make_pair(j,0);
      D[0][1]=std::make_pair(j,0);
   }

   // find max overlap satisfying the constraints
   px = 0;
   unsigned min_k = maxindel+1;
   for (unsigned i=minprefixoverlap+1; i <= P.size(); i++)
      for (unsigned k=0; k <= maxindel; k++) {
          unsigned ndiags = (A.size()-D[i][k].first-k);
          if (D[i][k].second / (1.0*ndiags) >= (int)minaveragequality &&
              k < min_k) {
            px = i-1;     
            min_k = k;
          }
      }

   for (unsigned i=0; i<=P.size();i++) {
       delete [] Dprev[i];
       delete [] D[i];
   }
   delete [] Dprev;
   delete [] D;

   // solving the symmetric case

   Dprev = new std::pair<int,int>*[A.size()+1]; 
   D = new std::pair<int,int>*[A.size()+1];        
   // initializing first column, and first row of second column
   for (unsigned i=0; i<=A.size();i++) {
       Dprev[i] = new std::pair<int,int>[maxindel+1];
       D[i] = new std::pair<int,int>[maxindel+1];
       for (unsigned k=0; k<=maxindel;k++) {
           if (i==k)  
              Dprev[i][k] = std::make_pair(0,0);
           else
              Dprev[i][k] = infinitypair;
           if (i==0)
              D[i][k] = std::make_pair(0,0);
       }
   }
   // filling "matrix" column-by-column
   for (unsigned j=1; j<=P.size(); j++) {
      for (unsigned i=1; i<=A.size(); i++)
         for (unsigned k=0; k<=maxindel; k++) {
            std::pair<int,int> left = infinitypair, diag = infinitypair, up = infinitypair;
            if (k>0) {
               left = Dprev[i][k-1];
               up = D[i-1][k-1];
            }
            // if P[j-1]==A[i-1] with probability (1 - error probability p) then matchscore = -10 log_10 p 
            // if P[j-1]!=A[i-1] then match probability is q=p/3 and matchscore = -10 log_10 (1-q)
#ifdef IGNORE_QUALITY
            int matchscore = 40;
#else
            int matchscore = int(P.getQuality()[j-1])-64.0; // from ascii to score          
#endif
            double p = pow(10.0, -matchscore/10.0);
            double q = p/3; 
            int mismatchscore = -10*log10(1-q);
            int diagonalscore = (P.getPattern()[j-1]==A[i-1]?matchscore:mismatchscore);
	    if (matchscore > (int)mismatchlimit) {
	       if (P.getPattern()[j-1]==A[i-1]) 
                  diag = Dprev[i-1][k];   
	       else if (k>0)
                  diag = Dprev[i-1][k-1];   	       
	    }
	    else 	
   	       diag = Dprev[i-1][k];  
            // take max of three scores
            int jj = left.first;
            int s = left.second;
            if (s<up.second) {
               jj = up.first;
               s = up.second;
            }
            if (s<diag.second+diagonalscore) {
               jj = diag.first;
               s = diag.second+diagonalscore;
            } 
            D[i][k] = std::make_pair(jj,s); 
         }         
      // copy D to Dprev
      for (unsigned i=1; i<=A.size(); i++)
         for (unsigned k=0; k<=maxindel; k++) 
            Dprev[i][k]=D[i][k];
      Dprev[0][0]=std::make_pair(j,0);
      D[0][1]=std::make_pair(j,0);
   }
   // find max overlap satisfying the constraints
   sx = P.size();
   min_k = maxindel+1;
   for (unsigned i=minsuffixoverlap; i <= A.size(); i++)
      for (unsigned k=0; k <= maxindel; k++) {
          unsigned ndiags = (P.size()-D[i][k].first+i-k)/2;
          if (D[i][k].second / (1.0*ndiags) >= (int)minaveragequality && D[i][k].first<(int)sx && k < min_k) {
             sx = D[i][k].first;             
             min_k = k;
          }
      }

   if (P.size()-sx<minsuffixoverlap)
      sx = P.size();
   for (unsigned i=0; i<=A.size();i++) {
       delete [] Dprev[i];
       delete [] D[i];
   }
   delete [] Dprev;
   delete [] D;

   return std::make_pair(px,sx);
}


/**
 * Definitions for parsing command line options
 */

void print_usage(char const *name)
{
    cerr << "usage: " << name <<endl
         << " -a (--adapter) string " << endl
         << " -i (--maxindel) int " << endl
         << " -q (--minaveragequality) int " << endl
         << " -m (--mismatchlimit) int " << endl	 	 
         << " -p (--minprefixoverlap) int " << endl
         << " -s (--minsuffixoverlap) int " << endl
         << " -P (--parallel) int " << endl
         << " filename (.fastq) " << endl;
}

int main(int argc, char **argv) 
{
    /**
     * Parse command line parameters
     */
    if (argc <= 1)
    {
        print_usage(argv[0]);
        return 0;
    }

    unsigned maxindel = 0;
    unsigned minaveragequality = 0;
    unsigned minprefixoverlap = 0;
    unsigned minsuffixoverlap = 0;
    unsigned numberofpatterns = 0;
    unsigned mismatchlimit = 0;
#ifdef PARALLEL_SUPPORT
    unsigned parallel = 1;
#endif
    string adapter = "";
    
    static struct option long_options[] =
        {
            {"maxindel",     required_argument,       0, 'i'},
            {"minaveragequality",     required_argument,       0, 'q'},
            {"minprefixoverlap",      required_argument, 0, 'p'},
            {"minsuffixoverlap",      required_argument, 0, 's'},
            {"adapter",      required_argument, 0, 'a'},
	    {"mismatchlimit",      required_argument, 0, 'm'},
            {"parallel",      required_argument, 0, 'P'},
            {0, 0, 0, 0}
        };
    int option_index = 0;
    int c;
    while ((c = getopt_long(argc, argv, "i:q:p:s:a:m:P:",
                                 long_options, &option_index)) != -1) 
    {
        switch(c) 
        {
        case 'i': 
            maxindel = atoi(optarg); break;
        case 'q':
            minaveragequality = atoi(optarg); break;
        case 'p':
            minprefixoverlap = atoi(optarg); break;
        case 's':
            minsuffixoverlap = atoi(optarg); break;
        case 'a':
            adapter = string(optarg); break;
        case 'm':
            mismatchlimit = atoi(optarg); break;
        case 'P':
#ifdef PARALLEL_SUPPORT
            parallel = atoi(optarg); break;
#else
            cerr << "readaligner: Parallel processing not currently available!" << endl 
                 << "Please recompile with parallel support; see README for more information." << endl;
            return 1;
#endif
        case '?':
            print_usage(argv[0]);
            return 1;
        default:
            print_usage(argv[0]);
            std::abort ();
        }
    }
    if (adapter == "")
    {
        cerr << "readaligner: missing adapter!" << endl;
        print_usage(argv[0]);
        return 1;
    }

    // Parse filename
    if (argc - optind != 1)
    {
        cerr << "readaligner: missing input files!" << endl;
        print_usage(argv[0]);
        return 1;
    }
     
    string patternfile;
    patternfile = string(argv[optind++]);   
    
    /**
     * Initialize shared input reader(s)
     */
    InputReader *inputr = 0;
    inputr = InputReader::build(InputReader::input_fastq, numberofpatterns, patternfile);
    assert(inputr != 0);

#ifdef PARALLEL_SUPPORT
    if (parallel != 0)
    {
       omp_set_num_threads(parallel);
    }
    else 
       cout << omp_get_max_threads() <<endl;
#pragma omp parallel 
#endif    
{
    /**
     * Main loop: iterate through input
     */
    Pattern pattern;
    std::pair<unsigned,unsigned> overlaps;
    while (inputr->next(pattern)) 
    {
        //pattern.truncate(truncate);
        overlaps = overlapAlign(adapter, pattern, maxindel, minaveragequality, minprefixoverlap,
                                minsuffixoverlap, mismatchlimit);  
        {
#ifdef PARALLEL_SUPPORT
#pragma omp critical (STDOUT)
#endif
            {
               cout << pattern.getId() << "," << overlaps.first << "," << overlaps.second << endl;
            }
        } // end of #pragma omp critical (STDOUT)
    }       
       
} // end of #pragma omp parallel

    
    delete inputr; 
}
